<?php

namespace Drupal\moodle_field_widget;

use Drupal\Component\Serialization\Exception\InvalidDataTypeException;
use Drupal\Component\Serialization\Json;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use GuzzleHttp\ClientInterface;
use GuzzleHttp\Exception\GuzzleException;
use Psr\Log\LoggerAwareInterface;
use Psr\Log\LoggerAwareTrait;

/**
 * Provides methods for interacting with the Moodle API.
 */
class MoodleService implements LoggerAwareInterface {

  use StringTranslationTrait;
  use LoggerAwareTrait;

  /**
   * The Moodle API URL.
   */
  protected const API_URL = 'https://placeholder/api/';

  /**
   * The moodle configuration.
   *
   * @var \Drupal\Core\Config\Config
   */
  protected $config;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The HTTP client.
   *
   * @var \GuzzleHttp\ClientInterface
   */
  protected $httpClient;

  /**
   * Constructs a MoodleService object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager service.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The config factory service.
   * @param \GuzzleHttp\ClientInterface $http_client
   *   The HTTP client.
   */
  public function __construct(
    EntityTypeManagerInterface $entity_type_manager,
    ConfigFactoryInterface $config_factory,
    ClientInterface $http_client,
  ) {
    $this->entityTypeManager = $entity_type_manager;
    $this->config = $config_factory->get('moodle_field_widget.settings');
    $this->httpClient = $http_client;
  }

  /**
   * Get default parameters that we need for an API request.
   *
   * @return array
   *   An array of key/value pairs.
   */
  protected function getApiRequestDefaultParams():array {
    return [
      'wstoken' => $this->config->get('api_token'),
      'moodlewsrestformat' => 'json',
    ];
  }

  /**
   * Make a Moodle API request.
   *
   * Sample usage:
   * ```php
   *   $this->makeApiRequest('core_course_get_courses_by_field', [
   *     'field' => 'category',
   *     'value' => 123,
   *   ]);
   * ```
   *
   * @param string $function
   *   The webservice function. E.g. `core_course_get_courses_by_field`.
   * @param array $params
   *   Additional parameters for the request. `wstoken` and `moodlewsrestformat` parameters
   *   are automatically populated.
   *
   * @return mixed
   *   Data returned from the request.
   */
  public function makeApiRequest(string $function, array $params = []): mixed {
    // Get the API url.
    if (!$this->config->get('api_url')) {
      // No API url, log an error.
      $this->logger->error($this->t('Unable to make a Moodle API request because no Moodle Webservice API URL has been configured.'));
      return FALSE;
    }

    $api_url = $this->config->get('api_url');

    // Gather all the parameters that need to be added to our query.
    $params = array_merge(
      $this->getApiRequestDefaultParams(),
      ['wsfunction' => $function],
      $params
    );

    // Concatenate our query parameters onto the URL.
    if ($params) {
      $query_parts = [];

      foreach ($params as $param_key => $param_value) {
        $query_parts[] = $param_key . '=' . $param_value;
      }

      $api_url .= '?' . implode('&', $query_parts);
    }

    // Send the request.
    try {
      // Make an API request using the HTTP client.
      /** @var \GuzzleHttp\Psr7\Response $response */
      $response = $this->httpClient->request('POST', $api_url);

      // Get the result.
      $result = Json::decode(
        $response->getBody()->getContents()
      );

      // Log an error if the response status code is not 200 (OK).
      if ($response->getStatusCode() != 200) {
        $this->logger->error('Request to the Moodle Webservice @endpoint endpoint received a not-OK response code: @code', [
          '@context' => $response->getStatusCode(),
          '@endpoint' => $api_url,
        ]);
      }

      // If the call returns an error message, log that.
      if ($result && isset($result['exception'])) {
        $this->logger->error('Request to the Moodle Webservice @endpoint endpoint returned an exception: @exception; message: @message', [
          '@endpoint' => $api_url,
          '@exception' => $result['exception'],
          '@message' => $result['message'] ?? '',
        ]);
      }
    }
    catch (GuzzleException $e) {
      $this->logger->error('Error performing HTTP request to Moodle Webservice @endpoint endpoint: @context', [
        '@context' => $e->getMessage(),
        '@endpoint' => $api_url,
      ]);
    }
    catch (InvalidDataTypeException $e) {
      $this->logger->error('Error decoding JSON returned by Moodle Webservice @endpoint endpoint: @context', [
        '@context' => $e->getMessage(),
        '@endpoint' => $api_url,
      ]);
    }
    catch (\Exception $e) {
      $this->logger->error('Unexpected exception performing request to Moodle Webservice @endpoint endpoint: @context', [
        '@context' => $e->getMessage(),
        '@endpoint' => $api_url,
      ]);
    }

    return $result;
  }

  /**
   * Get all courses from Moodle, optionally filtered by a value.
   *
   * @param string $category
   *   A category value with which to filter courses.
   *
   * @return array
   *   An array of courses.
   */
  public function getCourses(string $category):array {
    $params = [];

    if ($category) {
      $params = [
        'field' => 'category',
        'value' => $category,
      ];
    }

    $data = $this->makeApiRequest('core_course_get_courses_by_field', $params);

    // Return the courses.
    if (isset($data['courses'])) {
      return $data['courses'];
    }

    return [];
  }

}
